from ConfigParser import (
    ConfigParser,
    NoSectionError)
import errno
import os
import re
from StringIO import StringIO
import subprocess
import sys

from Cheetah.Template import Template

import hookenv


CHARM_DIR = os.getcwd()
HOME = '/home/ubuntu'
PROJECT_DIR = os.path.join(HOME, 'juju-reports')
INI = 'production.ini'


def in_juju_reports(fn):
    def dec_fn(*args):
        os.chdir(PROJECT_DIR)
        fn(*args)
        os.chdir(CHARM_DIR)
    return dec_fn


def get_ini_path():
    return os.path.join(PROJECT_DIR, INI)


def get_ini():
    config = ConfigParser()
    config.read(get_ini_path())
    return config


def update_or_install_ini(mongo_url=None):
    config = get_ini()
    if mongo_url is not None:
        config.set('app:main', 'mongo.url', mongo_url)
    config.write(open(get_ini_path(), 'w'))
    return config


def get_configured_port(ini):
    return ini_get(ini, 'server:main', 'port')


def set_port(ini, open_port=True):
    port = get_configured_port(ini)
    if port is None:
        return
    if open_port:
        hookenv.open_port(port)
    else:
        hookenv.close_port(port)


def call_script(script_path, root=False):
    env = os.environ.copy()
    ini = get_ini_path()
    env['HOME'] = HOME
    env['INI'] = ini
    if root:
        args = []
    else:
        args = ['su', 'ubuntu', '-p', '-c']

    args.append(script_path)
    subprocess.check_call(args, env=env, cwd=PROJECT_DIR)
    return env


def stop():
    try:
        call_script('scripts/stop', root=True)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def start():
    call_script('scripts/run')


def get_config():
    config = hookenv.config()
    if config['source'] == '':
        raise IncompleteConfig('source not specified.')
    if re.match('(lp:|.*://bazaar.launchpad.net/)', config['source']):
        if config['lp-key'] == '':
            raise IncompleteConfig('lp-key not specified.')
    return config


class IncompleteConfig(ValueError):
    """Error class indicating that ssh is not set to a valid value."""


@in_juju_reports
def update_source(source_url, revno, ssh_key):
    if ssh_key != '':
        fd = os.open('/root/.ssh/lp_rsa', os.O_WRONLY | os.O_TRUNC |
                     os.O_CREAT, 0600)
        try:
            os.write(fd, ssh_key)
        finally:
            os.close(fd)
        # Login to LP
        subprocess.check_call(['bzr', 'lp-login', 'juju-qa-bot'])

    # Grab the code
    revno = 'revno:%s' % revno
    bzr_cmd = ['bzr', 'pull', '--overwrite', source_url, '-r', revno, '-d',
               PROJECT_DIR]
    subprocess.check_call(bzr_cmd)

    # Remove the unnecessary INI files--if something tries to use them we need
    # a failure.
    for ini in ['test.ini', 'development.ini']:
        try:
            os.unlink(os.path.join(PROJECT_DIR, ini))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


@in_juju_reports
def install_production(dist_clean):
    subprocess.check_call(['make', 'deploydeps'])
    if dist_clean:
        subprocess.check_call(['make', 'distclean'])
    subprocess.check_call(['make', 'prodinstall'])
    subprocess.check_call(['apt-get', 'autoremove', '--yes'])


def install_cronjob(interval, error_email):
    t = Template(file='templates/crontab.tmpl')
    t.interval = interval
    if error_email is not None:
        t.error_email = 'MAILTO=%s' % error_email
    else:
        t.error_email = ''
    t.reports_home = PROJECT_DIR
    open('/etc/cron.d/ubuntu', 'w').write(str(t))


def ini_get(ini, section, key):
    try:
        return ini.get(section, key)
    except NoSectionError:
        return


def get_mongo_url():
    """Use the 'database' relation settings to build a mongodb url.

    If no mongodb units are present, return an empty string.
    """
    host_ports = []
    replset = None
    for relation_id in hookenv.relation_ids('database'):
        for unit in hookenv.related_units(relation_id):
            relation_data = hookenv.relation_get(rid=relation_id, unit=unit)
            # Consider configuration incomplete until port is supplied.
            if (relation_data is None or 'port' not in relation_data or
                'hostname' not in relation_data):
                continue
            if replset is None:
                replset = relation_data['replset']
            # All units should be members of the same replication set.
            elif replset != relation_data['replset']:
                raise AssertionError('DB instances are from different sets!')
            host_ports.append(
                '%(hostname)s:%(port)s' % relation_data)
    if len(host_ports) == 0:
        return ''
    return 'mongodb://%s/?replicaSet=%s' % (','.join(host_ports), replset)


def update_website(ini):
    relation_settings = {
        'port': get_configured_port(ini),
        'hostname': hookenv.unit_private_ip(),
    }
    for relation_id in hookenv.relation_ids('website'):
        hookenv.relation_set(relation_id, relation_settings)


def update_from_config():
    ini = get_ini()
    try:
        try:
            mongo_url = get_mongo_url()
            if mongo_url == '':
                raise IncompleteConfig('No mongodb set up.')
            config = get_config()
        except:
            set_port(ini, False)
            stop()
            raise
    except IncompleteConfig as e:
        subprocess.call(['juju-log', 'Incomplete config: %s' % e])
        sys.exit(0)
    set_port(ini, False)
    stop()
    update_source(config['source'], config['revno'], config['lp-key'])
    # ini may be updated by update_source
    ini = get_ini()
    install_production(config['dist-clean'])
    ini = update_or_install_ini(mongo_url)
    install_cronjob(config['cron-interval'], config['error-email'])
    start()
    set_port(ini, True)
    update_website(ini)
